(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/CodGenC.sml,v
   Comment: Common module for all C code generators. This is only general C stuff.
   Release: scope-0-16_final 

     Copyright (C) 2001-2003 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   Besides some typical C output tools, it contains a function to generate
   eexternal functions referenced in source code. The functions generated 
   are empty return void or return true, depending on type.
   
   CodGenC.sml,v 1.61 2003/07/23 22:08:23 wasowski Exp
*)

structure CodGenC = struct 

  open Haskellish Typeset Types CodGen
  infixr 0 $; 

  infix 5 THEN;
  infix 5 ELSE;
  infix 3 --;

  val revAppend = List.revAppend
  
  val padLeft = StringCvt.padLeft #" "

  (* implementation *)

  fun itoa (i :int) : string =
	if i < 0 then ("-" ^ Int.toString (~i)) else Int.toString i
  fun mitoa (is :int list) : string list =
	map itoa is

  fun dtoa (r :real) :string =
	if r < 0.0 then ("-" ^ Real.toString (~r)) else Real.toString r


  val NULL = 0
  val SNULL = "0"

  datatype ifclause = IfClause of codgen

  (* some lazy code generation. Call result with empty (or any releveant)
     output continuation to force generating the code. *)

  fun TYPEDEF (oldty :string) (newty :string) :codgen =
	cg ["typedef ", oldty, " ", newty, ";"]

  fun RETURN  (va : codgen) : codgen = sg "return (" -- va -- sg ");"
  fun RETURNS (cs : string) : codgen = cg ["return (",cs,");"]

  fun IF (cond : codgen) :ifclause = 
	IfClause (sg "if (" -- cond -- sg ") ")
  
  fun (IfClause ifc :ifclause) THEN (thc : codgen) :codgen = 
	ifc -- sg " { \n" -- thc -- sg ";}"

  fun CASE (ca :string) (body :codgen) = 
	cg ["case ", ca, ":\n"] -- body

  fun SWITCH (var :string) (body :codgen) =
	cg ["switch (", var, ") { \n"] -- body -- sg "\t}\n"

  fun (thc :codgen) ELSE (ec :codgen) :codgen =
	thc -- sg "else {\n " -- ec -- sg ";} "
	
  fun FUNC (header :codgen) (body :codgen) :codgen =
	sg "\n" -- header -- sg " {\n\n\t" -- body -- sg "\n}\n"

  fun HDRPS (rety :string) (name :string) (ps :string) :codgen =
	cg [rety, " ", name, "(", ps, ")"]

  fun HDR (rety :string) (name :string) :codgen =
	HDRPS rety name "VS_VOID"

  fun VAR_DECL (ty :string) (varname :string) :codgen = cg [ty, " ", varname]

  fun ARRAY_DECL (celltype :string) (varname :string) (size :string) :codgen =
	cg [celltype, " ", varname, "[", size, "]"]
  
  infix 4 INIT
  
  fun cg1 INIT cg2 :codgen = cg1 -- sg " = " -- cg2 -- sg ";\n"

  fun DEFINE (macro :string) (value :codgen) :codgen =
	cg ["#define ", macro, "\t" ] -- value -- eol

  fun UNDEF (macro :string) :codgen = cg ["#undef ", macro] -- eol

  (* generate a file include directive *)

  fun INCLUDE (fname :string) :codgen    = cg ["#include ", quote fname]

  (* generate a file include directive using only standard search path *)

  fun INCLUDESTD (fname :string) :codgen = cg ["#include ", angle fname] 
	
  (* the rest *)

  val arri_width = 9
  val int_width  = 5

  fun mkchunks (l :int) ([] : 'a list) = []
    | mkchunks l xs = 
	List.take (xs,l) :: mkchunks l (List.drop (xs,l)) 
	handle Subscript => [xs]

  (* build an array initialization block consisting of cells,
     formated cols cells per row, with comment on the left indicating
     the position in the array. cells are treated in stable way 
     (not reversed) *)

  fun mk_commented_arrini (cells :string list) (cols :int) :codgen =
	let val width    = size (itoa (length cells))
	    val chunks   = mkchunks cols cells
	    val offsets  = List.tabulate (length chunks,fn i => i*cols)
	    val f        = fn n => cg ["\n\t/* ",padl width (itoa n)," */ "]
	    val comments = map f offsets 
	    val cells    = map (fn c => cg (join_commas c [","])) chunks
	    val block    = ListPair.foldl (fn (co,ce,e) => e--co--ce) eoc 
				(comments,cells)
	in sg " {" -- block -- sg "\n}" end

  (* following functions are in some ways dual to presented above
     infixes version. These are more convenient to use in some situations. 
     TODO: I HOPE I WILL REMOVE THEM *)

  (* if size is NONE it is computed from cells assuming each cell contributing
     1 to size. Size is given as a string so you can give an empty string when
     there is no easy way to determine the size. However note that size was
     added here to make debugging easier (debuggers ofen have nicer display
     for arrays with known sizes). *)

   (* Are the following dumps neeedd here ? FIXME *)


   fun dump_value (Types.ValI i) :string = itoa i
     | dump_value (Types.ValR r) = Real.toString r
     | dump_value (Types.ValA lst) = 
	String.concat $ "{ " :: join_commas (map dump_value lst) [" }"]

   fun dump_const (Exp.CstI i) :string  = itoa i
     | dump_const (Exp.CstR r)		= dtoa r
     | dump_const (Exp.CstC c)		= String.implode [#"'",c,#"'"]

 (* printing expressions *)

fun dump_Prim2 Prim.OpAdd  = "+"
  | dump_Prim2 Prim.OpSub  = "-"
  | dump_Prim2 Prim.OpMul  = "*"
  | dump_Prim2 Prim.OpMod  = "%"
  | dump_Prim2 Prim.OpDiv  = "/"
  | dump_Prim2 Prim.OpLt   = "<"
  | dump_Prim2 Prim.OpLte  = "<="
  | dump_Prim2 Prim.OpGt   = ">"
  | dump_Prim2 Prim.OpGte  = ">="
  | dump_Prim2 Prim.OpEq   = "=="
  | dump_Prim2 Prim.OpNeq  = "!="
  | dump_Prim2 Prim.OpBXor = "^"
  | dump_Prim2 Prim.OpAnd  = "&&"
  | dump_Prim2 Prim.OpOr   = "||"
  | dump_Prim2 Prim.OpBAnd = "&"
  | dump_Prim2 Prim.OpBOr  = "|"
  | dump_Prim2 Prim.OpShr  = ">>"
  | dump_Prim2 Prim.OpShl  = "<<"

fun dump_Prim1 Prim.OpNot   = "!"
  | dump_Prim1 Prim.OpBNot  = "~"
  | dump_Prim1 Prim.OpMinus = "-"

fun csizeof( ty :simplety ) :string = "sizeof(" ^ tyname ty ^ ")"

end
