(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/CodGenCF_front.sml,v
   Comment: Front end of CodGenCF code generator. Takes a flat.sml model and
	    compiles it to flat lrsystem (intermediate representation)
   Release: scope-0-16_final 

     Copyright (C) 2003 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   NOTES:

   This modules tries to avoid operating on numbers (addresses, references,
   etc). Symbolic names are used as far as possible and actual array
   addresses are assigned in scheduler. One exception is actions and guards,
   as we need to count the number of unique before type inference. This is
   conveniently done by constructing an ulist, which performs a scheduling in
   place.

   CodGenCF_front.sml,v 1.9 2003/07/29 12:44:08 wasowski Exp
*)

structure CodGenCF_front = struct
(* some interface *)

open List Flat CodGenCF_LR CodGenC_front 
infix 3 --;

fun clean_tgts (a :anatomy) (tgts :symbol list) :symbol list =
	List.filter (fn t => SE.defined a t) tgts

(* compile a single transition, suitable for folding *)

fun comp_tran (a :anatomy) (ets :evtype env) 
	(t :tran, (tenv :tenv, guards, actions :actions)) 
		:tenv * guards * actions = 
	let val subst = comp_subst ets (getebind (#trigger t))
	    val (actref,actions1) = 
			alloc (subst_action subst (#action t),actions)
	    val guard = if null (#guard t) then always else (#guard t)
	    val (guardref,guards1) =
			alloc (subst_guard subst guard, guards)
            val targets= map LRAndStateRef (clean_tgts a (#targets t))
	    val trans  = SE.lookup tenv (#1 (getebind (#trigger t)))
				handle NotFound => [] 
	    val trans1 = List.revAppend (targets, LRMrk STMRK::trans)
	    val trans2 = LRAcGd (actref+markno,guardref+markno) ::trans1
	    val trans3 = map LRAndStateRef (#pos t @ #neg t) @ trans2
	    val trans4 = LRCondCounters(length (#pos t), length (#neg t))::trans3
	    val tenv1  = SE.bind1 tenv (#1 (getebind (#trigger t)), trans4)
	in (tenv1, guards1, actions1) end
	handle NotFound => impossible "#200307211" 

(* compile a transitions list (already degrouped, no dynmic scopes) *)

fun comp_trans (a :anatomy) (ets: evtype env) (trans : tran list) 
			:tenv * guards * actions = 
	let val guards  = #2 (alloc (always,UList.initialize guardsize))
	    val actions = UList.initialize actionsize
	in foldl (comp_tran a ets) (empty,guards,actions) trans end
	handle NotFound => impossible "#200307210" 

(* build current configuration vector *)

fun comp_machines (ms :machineid list) :lratom list =
	foldr (fn (m,e)=>LRStateLabel m::LRMrk STMRK::e) [] ms

(* compile anatomy: return dictionary state -> machine, not-reversed *)

fun comp_anatomy (a :anatomy) :lratom list * int =
	let val pairs = SE.toList a
	    fun add ((ast,ost),e) =
		LRStateLabel ast :: LROrStateRef ost :: e
	    val anatomy = tabulate (markno,fn _=> LRMrk STMRK) 
				@ foldr add [] pairs
	in (anatomy,length pairs) end

(* build initialization transition *)

fun comp_init (a :anatomy) (ms :machineid list) ((ac,ts) :action * stateid list) 
		:action * lratom list =
	let val ts = clean_tgts a ts
	    fun get (m,e) = case (List.find (fn t=>SE.lookup a t = m) ts) of
				  SOME s => LRAndStateRef s ::e
				| NONE   => impossible "#200307290" 
			(* flattening should return complete iniconf *)
	in (ac,foldr get [] ms) end


(* compile system, compute some sizes and infer types *)

fun comp_system (queuelen :int) (sys : system) :lrsystem =
	let val (anatomy,states)= comp_anatomy (#anatomy sys)
	    val (tenv,guards,actions)= comp_trans (#anatomy sys) 
					(#evtypes sys) (#trans sys) 
	in { name     = Symbol.name (#name sys),
	     fname    = Symbol.name (#fname sys),
	     anatomy  = anatomy,
	     mconf    = comp_machines (#machines sys),
	     tenv     = tenv,
	     init     = comp_init (#anatomy sys) (#machines sys) (#init sys),
	     
	     signals  = comp_signals (#2(UList.alloc (#1(#init sys),actions))),
	     actions  = actions,
	     guards   = guards,

	     evtypes  = #evtypes sys,
	     vartypes = #vartypes sys,
	     venv     = #venv sys,
	     cenv     = #cenv sys,

	     queuelen = comp_queuelen queuelen (#evtypes sys),
	     statew   = length (#machines sys),
	     states   = states }
	end

(* infer type sizes for given system in LR *)

fun resolve_types (sys :lrsystem) :lrsizes = 
	let val evnumber  = SE.count (#evtypes sys)
	    val signumber = UList.length (#signals sys)
	in { 
		actionref = mintysize (UList.length(#actions sys) + markno -1),
		guardref  = mintysize (UList.length(#guards  sys) + markno -1),
		signalref = mintysize (evnumber + 1 + signumber - 1),
		(* signalrefsize can be done better, substracting 
		   the length of the last entry *)
		andstref  = mintysize (#states sys + 1 - 1),
		orstref   = mintysize (#statew sys - 1),
		evnumber  = evnumber 
	} end


end
