(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/Simplifier.sml,v
   Comment: Positive and negative condition simplifier.
   Release: scope-0-16_final

     Copyright (C) 2001-2003 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   NOTES:
   
   I do not claim that this algorithm gives a minimal solution. The only
   thing it should do is to return a solution which is not bigger than
   initial condition and has the same semantics (is equivalent in given
   model).

   The module also contains functions to call recursively poscond/negcond
   simplificator for all transitions in all systems in the module. Note that
   the name is misleading a bit. It does not simplify the model, it only
   tries to eliminate redundant conditions form transitions. Warnings are
   not issued in whole model mode. This is because it is unclear if
   redundancy should be considered an error or an attempt to aid code
   generator.

   Simplifier.sml,v 1.15 2003/07/07 23:53:15 wasowski Exp
*)

structure Simplifier = struct

open Absyn 
infixr 0 $
infix 9 sub

val exists = List.exists
val all = List.all

(* normalise (pump) as many negative states to 
   positive states as possible. *)

fun neg2pos (h :'aor hierarchy) (pos,neg)
		:'aor andstate list * 'aor andstate list =
	let val npos = map nameand pos
	    fun f (n,(pos,neg)) =
		if exists (fn p=> parentand n = p) npos then
			case siblingandst h n of
				[s1,s2] => if nameand n = nameand s1 then (s2::pos,neg) 
						     else (s1::pos,neg)
				|     _ => (pos, n::neg)
		else (pos, n::neg)
	in foldl f (pos,[]) neg end

(* guarantee uniqueness and propagate positive conditions *)

fun proppos (h :'aor hierarchy) (pos :'aor andstate list) :stateid list =
	let val npos = map nameand pos
	    val npos = foldl (fn(p,e)=>if all (fn xe=>xe<>p) e then p::e else e) 
				[] npos
	    fun f [] = []
	      | f (s1::ss) = if exists (ancestorh h s1) npos 
			     then f ss else s1::f ss
	in f npos end

(* guarantee uniqueness and propagate negative conditions *)

fun propneg (h :'aor hierarchy) (neg :'aor andstate list) :stateid list =
	let val nneg = map nameand neg
	    val nneg = foldl (fn(p,e)=>if all (fn xe=>xe<>p) e then p::e else e) 
				[] nneg
	    fun f [] = []
	      | f (s1::ss) = if exists (fn s2=>ancestorh h s2 s1) nneg
				then f ss else s1::f ss
	in f nneg end


(* remove negative conditions implied by trivial positive conditions -
   only single positive conditions are taken into account, but what does
   it really change ? *)	

fun rmneg (h :'aor hierarchy) (pos :stateid list, neg :stateid list) 
	:stateid list =
     let fun f (n :stateid) (p :stateid) :bool = 
		ancestralh h n p orelse orthogonal h n p
     in List.filter (fn n => all (f n) pos) neg end


(* this is a dumb simplifier - it takes all four algorithms present and
   applies them in sequence *)

fun simplify (h :'aor hierarchy) (pos :stateid list, neg :stateid list)
				:stateid list * stateid list =
     let val pos = List.map (valOf o getand o getstate h) pos
         val neg = List.map (valOf o getand o getstate h) neg
         val (pos1,neg1) = neg2pos h (pos,neg)
         val pos2 = proppos h pos1
         val neg2 = propneg h neg1
         val neg3 = rmneg h (pos2,neg2)
     in (pos2, neg3) end

(* simplify with messages *)

fun printcond (pos :stateid list, neg :stateid list) :string =
     let val spos = foldl (fn (s,e)=> s ^ " " ^ e) "" (map Symbol.name pos)
         val sneg = foldl (fn (s,e)=> s ^ " " ^ e) "" (map Symbol.name neg)
     in "[" ^ spos ^ "] AND NOT [" ^ sneg ^ "]" end

fun simplifyw (h :'aor hierarchy) (pos :stateid list, neg :stateid list)
					:stateid list * stateid list =
	let val (pos1,neg1) = simplify h (pos,neg)
	in if length pos1 + length neg1 < length pos + length neg then
		(Err.warning "Redundant condition found in transition";
		 Err.warning $ "Condition: " ^ printcond (pos,neg) ^ 
			      " simplified to: " ^ printcond (pos1,neg1))
	   else ();
	   (pos1,neg1)
	end


(* simplify a transition with warnings! 
 * should only be used for testing not for production - the same for
 * following. It produces unwanted messages 
 *)

fun simplify_tran (h :'aor hierarchy) (t :'atr tran) :'atr tran =
     let val (pos,neg) = simplifyw h (#pos t,#neg t)
     in { trigger = #trigger t, scope = #scope t, pos = pos, neg = neg,
          guard= #guard t, action = #action t, targets = #targets t,
	  annotation= #annotation t }
     end handle e => intErr e 200212110

(* simplify system *)

fun simplify_sys (sys :('aor,'atr) system) :('aor,'atr) system =
     { name = #name sys, fname = #fname sys, 
       hierarchy = #hierarchy sys, 
       vartypes = #vartypes sys, evtypes = #evtypes sys, 
       evgroups = #evgroups sys, otypes = #otypes sys,
       denv = #denv sys, venv = #venv sys, cenv = #cenv sys,
       trans = map (simplify_tran (#hierarchy sys)) (#trans sys)
     }

(* simplify model *)

fun simplify_model m = map simplify_sys m
	 

   
	

		
	

end
