(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/Stat.sml,v
   Comment: Some simple statistics for Absyn models
   Release: scope-0-16_final

     Copyright (C) 2001-2003 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   NOTES:

   The simplicty of this statistics is among the other the fact that no
   boolean logics (BDDs) is applied here.

   Stat.sml,v 1.29 2003/07/11 15:20:55 wasowski Exp
*)

structure Stat = struct

(* interface to other modules *)

open Absyn;
infixr 0 $;

structure SE = SymbolEnv

val all = List.all

(* implementation *)

fun maxl l = foldl (fn (x,e) => if x>e then x else e) 0 l
fun suml l = foldl (fn (x,e) => x+e) 0 l

fun depthor  (ORState  st) = 1 + (maxl $ map depthand $ #children $ st)
and depthand (ANDState st) = 1 + (maxl $ map depthor	$ #children $ st)

fun depth (sys :('aor,'atr) system) :int = depthand $ gettree $ #hierarchy $ sys

fun trans (sys :('aor,'atr) system) :int = length $ #trans $ sys

(* returns the number of andstates *) 

fun and_states (sys :('aor,'atr) system) =
      hfoldand #2 (fn (_,c) => c+1) 0 (gettree $ #hierarchy sys)

(* lists all andstates *)

fun list_andst (sys :('aor,'atr) system) :stateid list =
	hfoldand #2 (fn (s,ls)=> nameand s::ls) [] (gettree $ #hierarchy sys)

(* returns the number of orstates *)

fun or_states (sys :('aor,'atr) system) = 
        hfoldand (fn (_,c) => c+1) #2 0 (gettree $ #hierarchy sys)

(* lists all or states *)

fun list_orst (sys :('aor,'atr) system) :stateid list =
	hfoldand (fn (s,ls)=> nameor s::ls) #2 [] (gettree $ #hierarchy sys)

(* returns the distribution of state types *)

fun state_dist (sys :('aor,'atr) system) :int * int = 
      hfoldand (fn (_,(oo,aa)) => (oo+1,aa)) (fn (_,(oo,aa))=> (oo,aa+1))
	(0,0) (gettree $ #hierarchy sys)

(* returns the number of all states in the system *)

fun states (sys :('aor,'atr) system) :int = 
      let fun f (_,count) = count + 1
      in hfoldand f f 0 (gettree $ #hierarchy sys) end

(* returns the number of basic states *)	

fun basic_states (sys :('aor,'atr) system) :int =
      let fun fa (s :'aor andstate,count) = if isbasicand s then count+1 else count
      in hfoldand #2 fa 0 (gettree $ #hierarchy sys) end

(* returns the number of transitions in the system *)

fun ntran (sys :('aor,'atr) system) :int = length (#trans sys)

(* number of transitions without actions (on tran) *)

fun ntran_noac (sys :('aor,'atr) system) :int =
	let fun count (t,e) = if (#action t = empty_action) then e + 1 else e
	in foldl count 0 (#trans sys) end

(* number of transitions without guard *)

fun ntran_nogd (sys :('aor,'atr) system) :int =
	let fun count (t,e) = 
		if (#guard t = [] orelse #guard t = [Cst (CstI 1)]) then e + 1
		else e 
	in foldl count 0 (#trans sys) end

(* number of transitions with action and guard *)

fun ntran_acgd (sys :('aor,'atr) system) :int =
	let fun count (t,e) =
		if #guard t <> nil andalso #guard t <> [Cst (CstI 1)] 
		    andalso #action t <> empty_action then e + 1 else e
	in foldl count 0 (#trans sys) end

(* number of transitions having negative condition *)

fun ntran_neg (sys :('aor,'atr) system) :int =
	let fun count (t,e) = if #neg t <> nil then e+1 else e
	in foldl count 0 (#trans sys) end

(* compute bound for state configuration cardinality *)

local 
      fun widthand (ANDState s) =					   
            Int.max(1,foldl (fn (s,e) => e + widthor s) 0 (#children s))
								   
      and widthor (ORState s) =
            foldl (fn (s,e) => Int.max(widthand s, e)) 0 (#children s) 
in 
      fun conf_card_bound (sys :('aor,'atr) system) :int = 
	widthand $ gettree $ #hierarchy sys
end

(* compute the number of redundant rors. Should return 0 for normalised
   models. Cryterium is more conservative than it could be. With not so
   much effor one can check if some target actions actually have a fixed
   scope because of positive condition or does not have unwanted scope
   of negative condition.  *)

fun rors (sys :('aor,'atr) system) =
      let val ts = #trans sys
          fun not_scoped s = all (fn t => #scope t <> nameor s) ts
          fun not_targeted s = all (fn t => all (fn tgt => tgt <> nameand s) 
					(#targets t)) ts
          fun fo (s,c) = 
		case (childrenor s) of 
		  [s1] => if (not_scoped s andalso not_targeted s1) 
			  then c+1 else c
		| _   => c
          fun fa (s,c) = c
      in hfoldand fo fa 0 (gettree $ #hierarchy sys) end

(* compute number of states having entry actions and number of states
   having exit actions *)

fun count_enex ((root,_) :'aor hierarchy) :int * int =
      let fun hasen s = if #entry s = empty_action then 0 else 1
          fun hasex s = if #exit  s = empty_action then 0 else 1
          fun fo (ORState s,e) = e
          fun fa (ANDState s,(en,ex)) = (en + hasen s, ex + hasex s)
      in hfoldand fo fa (0,0) root end 

(* compute number of entry/exit actions containing signals (counting number
   of states again) *)

fun count_sigenex ((root,_) :'aor hierarchy) :int * int =
      let fun hasensig s = if null (signals (#entry s)) then 0 else 1
          fun hasexsig s = if null (signals (#exit s))  then 0 else 1
          fun fo (ORState s,e) = e
          fun fa (ANDState s,(en,ex)) = (en + hasensig s, ex + hasexsig s)
      in hfoldand fo fa (0,0) root end

(* compute the maximum of size of target set on single entering (under
   single scope). Works on statically scoped transitions. *)
      
fun targets_card_bound (trans :static tran list) :int =
	let fun f ({annotation,...},e) = 
		foldl Int.max e (map (length o #tgts) annotation)
	in foldl f 0 trans end

(* leave only roots of forest tree given by state list *)
(* if given list is unique then results are unique *)

fun roots (h :'aor hierarchy) (ss :'aor state list) =
	let val names = map name ss
	    fun isroot s1 = List.all (fn s => not (ancestorh h s s1)) names
	in List.filter (fn s => isroot (name s)) ss end
	
(* compute list of scopes *)

fun get_scopes (sys :('aor,'atr) system) :stateid list =
	let val trans = Normalise.degroup_all (#evgroups sys) (#trans sys) 
	    val trans = StaticScope.ss_trans (#hierarchy sys) trans
	    val nonunique = List.concat (map (map #stscope o #annotation) trans)
	    val unique    = Set.addList (Set.empty, nonunique)
	in Set.listItems unique end

(* return number of history states *)

fun history_sts (sys :('aor,'atr) system) :int =
	let fun fo (ORState s, n)  :int = 
			if #marking s = History then n+1 else n
	    fun fa (ANDState s, n) :int = n
	in hfoldand fo fa 0 (gettree (#hierarchy sys)) end

(* check if state is exit-pure *)

fun is_expureand (ANDState s) :bool =
	(#exit s = empty_action) 
	andalso List.all (not o ishistoryor) (#children s)
	andalso List.all is_expureor (#children s)
	
and is_expureor (ORState s) :bool =
	List.all is_expureand (#children s)

(* compute number of events/signals, returns (|events|,|signals|) *)

fun eventno (sys :('aor,'atr) system) : int * int =
	let val lst = map #2 (SE.toList (#evtypes sys))
	in 
		foldl (fn ((i,_),(e,s)) => if i then (e,s+1) else (e+1,s)) 
			(0,0) lst 
	end

end
