(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/StaticScope.sml,v
   Comment: Machinery to compute static scope for transitions. If needed
	    transition is multiplied to eliminate dynamic scope.
   Release: scope-0-16_final

     Copyright (C) 2001-2003 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   NOTES:

   The module assumes that there are only event transitions coming in. GROUP
   TRIGGERED TRANSITION SHOULD BE FIRST MULTIPLIED WITH Normalise.degroup.

   1. scope of force state action to state which is already active is not
   correct. Behaves as self loop. Actually this is inconsistent in VS. The
   explicit target behaves as self loop while force state targets have
   internal scope. Yuck.

   StaticScope.sml,v 1.33 2003/07/24 07:45:06 wasowski Exp
*)

structure StaticScope = struct

open Absyn AbsynLogic WuddySys
infixr 0 $
infix 9 sub

infix 7 AND
infix 6 OR
infix 4 IMP BIIMP



(* compute the minLCA - lowest state being ancestor to at least one
   positive condition and all the targets of transition *)

fun mLCA (h :'aor hierarchy) (t :'atr tran) :'aor andstate = 
	let 
	    val ss = map (fn p => irrLCA h (p :: #targets t)) (#pos t)
	    fun f (s,mlca) = if ancestorh h s mlca then mlca else s
	    val mlca = foldl f (hd ss) (tl ss) 
		       handle Empty => impossible "tran should have 1+ poscond"
	    val mlca1 = getstate h mlca
	    val mlca2 = if isor mlca1 then getstate h $ parent mlca1 
			else mlca1
	in valOf $ getand mlca2 end handle e => intErr e 200212111

(* build bdd of part of the statechart below minLCA, restrict it for
   positive and negative conditions and existentially quantify on
   nonrelevant variables.  *)

fun solve (h :'aor hierarchy) (t :'atr tran, mlca :'aor andstate) 
		:(stateid list * stateid list) list = 
	let 
	    val _    = sayln VHigh ("mLCA: " ^ Symbol.name (nameand mlca))
	    val r    = WuddySys.build (InclAND mlca)
	    val _    = sayln Verbose (WuddySys.sprMapping())
	    val pos1 = List.filter (ancestorhst h (nameand mlca)) (#pos t)
	    val neg1 = List.filter (ancestorhst h (nameand mlca)) (#neg t)
	  (*val r1   = WuddySys.IMP(r,build_guard (pos1,neg1))*)
	    val r1   = AbsynLogic.force r (pos1,neg1)
	    val ancs = List.filter (not o ancestorhst h (#scope t)) 
				$ List.filter (ancestorhst h $ nameand mlca)
				$ List.filter (isandh h) 
				$ List.concat 
				$ map (ancesth h) (#targets t)
	    val _    = sayln Verbose ("Number of targets in restriction step: " ^
				    Int.toString (length (#targets t)))
	    val _    = sayln Verbose ("Number of states we look at: " ^
				    Int.toString (length ancs))
	    val r2   = WuddySys.restrict_to r1 ancs
	in WuddySys.allsat r2 ancs end handle e => intErr e 200212112


(* Split targets of transition into ancestrally related groups with
   single scope. Note static source is not filled in. *)
     
fun groupByScope ([] :(stateid * stateid) list) :static = []
  | groupByScope ((a1,o1)::ta :(stateid * stateid) list) :static = 
	let val (grp,rest) = List.partition (fn (a2,o2) => o1 = o2) ta
	in {stsrcs=NONE, stscope=o1, tgts=a1::map #1 grp}
		:: groupByScope rest 
	end

(* make new scope for single target. pos is assumed to be not simplified. 
   Should be not less than what is returned from bdd pacakge.  It is then 
   suffcient to find the nearest ancestor on list of positive conditions. 
   returns target and its scope.
     
   es - explicit scope of transition. Used for uncovering algorithm *)

fun staticscope (h :'aor hierarchy) (pos :stateid list) 
			(tgt :stateid) :stateid * stateid =
	let val root = name (getroot h)
	    fun lowest (p,e) = 
		if ancestorh h p tgt then if ancestorh h e p then p else e else e
	    val s    = foldl lowest root pos
	    val ors  = map nameor $ childrenand (valOf $ getand $ getstate h s)
	    val sc   = valOf $ List.find (fn s => ancestorh h s tgt) ors
	    val _    = sayln Verbose ("Static scope found: "^ Symbol.name sc)
	in (tgt, name $ getstate h sc) end 
	
(* Append newly computed pos and negconds and voila.
   you have a static transition. cleanup propagation should be done *)
     
fun mk_stran (h :'aor hierarchy) (t :'atr tran) [(pos,neg)] : static tran list =
  
	(* do not add anything to conditions if there is only a single
	   solution. A simple workaround. General should do in all cases:
	   never add a condition if it is the same in all new cases. FIXME *)

	(let val (pos1,neg1) = Simplifier.simplify h (#pos t, #neg t)
	     val ts          = map (staticscope h pos) (#targets t)
	in [{ trigger= #trigger t, pos= pos1, neg= neg1,
	      guard= #guard t, action= #action t, scope= #scope t, 
	      targets= map #1 ts, annotation= groupByScope ts }]
	end handle e => intErr e 200212121)

    | mk_stran (h :'aor hierarchy) (t :'atr tran) soluts =
	(let fun f (pos,neg) =
		let val (pos1,neg1) = (pos @ #pos t, neg @ #neg t)
		    val (pos2,neg2) = Simplifier.simplify h (pos1,neg1)
		    val ts = map (staticscope h pos) (#targets t)
		in
		   { trigger= #trigger t,
		     pos= pos2, neg= neg2, scope= #scope t,
		     guard= #guard t, action= #action t,
		     targets= map #1 ts, annotation=groupByScope ts } 
		end
	in map f soluts end handle e => intErr e 200212120)

(* If transition is not to be fired ever it is eliminated. Otherwise static
   scope is computed (multiplying transitions if needed) and list of
   equivalent static transitions is returned.  *)

fun ss_tran (h :'aor hierarchy) (t :'atr tran) :static tran list = 
	let val t1     = Simplifier.simplify_tran h t
	    val mlca   = mLCA h t1
	    val soluts = solve h (t1,mlca)
	    val _      = sayln Verbose "Solutions BEGIN"
	    val _      = IfVerbose(fn () => app (sayln Verbose o sprSolution) soluts)
	    val _      = sayln Verbose "Solutions END"
	in mk_stran h t soluts end handle e => intErr e 200212102

  (* compute static scopes for all transitions in one shot *)

fun ss_trans (h :'aor hierarchy) (trans :'atr tran list) 
		:static tran list =
	List.concat $ map (ss_tran h) trans
	handle e => intErr e 200212101

fun ss_sys (sys :('aor,'atr) system) :('aor,static) system =
	settrans sys (ss_trans (#hierarchy sys) (#trans sys))

fun ss_model (m :('aor,'atr) model) : ('aor,static) model = map ss_sys m

(* computes minimal sources for given static transition *)
(* these are also called static or proper sources       *)

fun minsrcs_tran (phi_h :W.bdd, h:'aor hierarchy) (tran :static tran) :static tran = 
	let val psi_h = phi_h AND build_guard (#pos tran,#neg tran)
	    fun mkor (ORState s) :stateid list =
		foldl mkand [] (#children s)
	    and mkand (ANDState s,cont) :stateid list = 
		if always (psi_h IMP bddVar (#name s)) then
			let val stsrcs = List.concat (map mkor (#children s))
			in if null stsrcs then (#name s)::cont 
			   else stsrcs @ cont
			end
		else cont
	    fun mkanno {stsrcs,stscope,tgts} = 
		{ stsrcs = SOME (mkor (getorh h stscope)),
		  stscope= stscope, tgts= tgts }
	in setannotationtr tran (map mkanno (#annotation tran)) end 

fun minsrcs_trans (phi_h :W.bdd, h:'aor hierarchy) (trans :static tran list)=
	map (minsrcs_tran (phi_h,h)) trans


(* diagnostic/debugging/testing support *)

local 

      open AbsynPrint Typeset
in

      fun sprTgts ({tgts,stscope,stsrcs}) :string =
	String.concat ("[" :: S.name stscope :: "] " 
		:: join_blanks(map S.name tgts) [" "])

      fun sprSTran ({trigger,pos,neg,guard,action,targets,scope,annotation} 
		:static tran) :string =
	String.concat ( sprEBind (getebind trigger) :: " : POS"
		:: c_params(map S.name pos) (" NEG"
			:: c_params(map S.name neg) ( " -> "
				:: map sprTgts annotation )))
      
      fun prSTranln stran = println (sprSTran stran)
end
 

end
