(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/parserl.sml,v
   Comment: Parser for RL(restriction language)
   Release: scope-0-16_final

     Copyright (C) 2003 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   PROBLEM: only simple reals allowed, i.e [0-9]+\.[0-9]+. I was to lazy to
   write the grammar for all possible representations and I could not figure
   out how to use scan Real.scan and scan Int.scan to force integers whenever
   there is no dot and vice versa. This partly stems from the idea of
   complete separation of parser and lexer. This would be much easier to do
   with standard primitives if parser was able to rewind lexer (and it would
   if both were actually one big parser of two levels). Still it is unclear
   for me chow the two approaches differ in reporting errors.

   FIXME[EFFICIENCY]: this is wrong at the moment that lexer needs a string
   (the whole file needs to read in) and then it produces the list (list of
   tokens for entire file) which then is parsed. This would be ok in a lazy
   language (well... I do not know if strings can be lazy in any language),
   but uses far to much memory in this case. Some way of lazy compositions
   should be possible. (NO TIME AT THE MOMENT).


   parserl.sml,v 1.6 2003/05/08 09:26:27 wasowski Exp
*)

structure ParseRL : PARSERL = struct

open RL Parsercomb Types Err



infix 6 $-- --$ #-- --#
infix 5 --
infix 3 >> >>*
infix 2 >>=
infix 0 ||

(* lexical tokens *)

datatype token  = T_RESTRICTION | T_LBR | T_RBR | T_RESTRICTS | T_CONST
		| T_EQUALS | T_IMPOSSIBLE | T_OCCURS | T_PURE | T_DEAD 
		| T_COMMA | T_COLON | T_PARENTHESIZED | T_SEMICOLON 
		| T_CSTI of int  | T_CSTR of real | T_NAME of string

(* Scanner definition *)

(* helpers for semantic actions *)

fun isAlphaNumUndsc c = (Char.isAlphaNum c) orelse (c = #"_")
fun isAlphaUndsc    c = (Char.isAlpha c   ) orelse (c = #"_")

fun keyword "restriction"	= T_RESTRICTION   
  | keyword "restricts"		= T_RESTRICTS	
  | keyword "const"		= T_CONST
  | keyword "impossible"	= T_IMPOSSIBLE
  | keyword "occurs"		= T_OCCURS	
  | keyword "pure"		= T_PURE	
  | keyword "dead"		= T_DEAD	
  | keyword  name		= T_NAME name

fun notRPar c = c <> #")" 
fun notAsterisk c = c <> #"*"
fun notSlash c = c <> #"/"

fun mkreal (r,l) = T_CSTR (valOf (Real.fromString 
		(Int.toString r ^ "." ^ Int.toString l)))

(* productions *)

val scan_id = 
	getElem isAlphaUndsc -- getElems0 isAlphaNumUndsc 
						>> (implode o op::)

val scan_int  = 
	scan (Int.scan StringCvt.DEC)		

val scan_const = 
	   scan_int --$ "." -- scan_int >> mkreal
	|| scan_int >> T_CSTI 

val scan_token = 
	   scan_id				>> keyword
	|| scan_const
	|| "(" $-- getChars0 notRPar --$ ")"	>> (fn _ =>T_PARENTHESIZED)
	|| $ ","				>> (fn _ =>T_COMMA)
	|| $ "="				>> (fn _ =>T_EQUALS)
	|| $ ":"				>> (fn _ =>T_COLON)
	|| $ ";"				>> (fn _ =>T_SEMICOLON)
	|| $ "{"				>> (fn _ =>T_LBR)
	|| $ "}"				>> (fn _ =>T_RBR)

fun endComment _ =
	getChars0 notAsterisk #-- ($ "*/" || getItem >>= endComment) 

val skipComment = 
	("/*" $-- endComment #"_") >> (fn _ => NONE)

fun focus phrase _ = skipWS (phrase || skipComment >>= focus phrase)

val tokenizer : (char, token list) parser = 
	commitChar "legal token" (repeat0 (focus scan_token NONE)) 
		--# focus (eof()) NONE

(* Parser definition *)

(* type of semantic values *)

type 'a semval = (token,'a) parser

(* helper functions *)

fun get_name (T_NAME n) = n
  | get_name _ = raise Fail "#200303070" (* impossible *)
fun get_int  (T_CSTI i) = i
  | get_int  _ = raise Fail "#200303080" (* impossible *)
fun get_real (T_CSTR r) = r
  | get_real _ = raise Fail "#200303081" (* impossible *)

(* productions *)

val getName :string semval = 
	getElem (fn t => case t of T_NAME _ => true | _ => false)
						>> get_name

val getInt :int semval = 
	getElem (fn t => case t of T_CSTI _ => true | _ => false) 
						>> get_int

val getReal :real semval = 
	getElem (fn t => case t of T_CSTR _ => true | _ => false) 
						>> get_real

fun commasep (phrase :'a semval) :'a list semval = 
	repeat0 (phrase --# getLit T_COMMA) -- phrase 
						>> (fn (li,la) => li @ [la])

val variable :(simplety * string) semval = 
	getName -- getName			>> (fn (t,n) => (namety t,n))

val function :(simplety * string) semval = 
	getName -- getName --# getLit T_PARENTHESIZED  
						>> (fn (t,n) => (namety t,n))

val constant :value semval = 
	   getInt				>> ValI 
	|| getReal				>> ValR

val namelist :string list semval = 
	commasep getName 

val eventlist :property list semval = 
	commasep (getName --# getLit T_PARENTHESIZED)	>> map NeverE

val constvar :property semval = 
	variable --# getLit T_EQUALS --constant	>> (fn((t,n),v)=>ConstrV(t,n,v))

val constfun :property semval = 
	function --# getLit T_EQUALS --constant	>> (fn((t,n),v)=>ConstrF(t,n,v))

val constlist :property list semval = 
	commasep (constvar || constfun)

val varlist :property list semval = 
	commasep variable			>> map Dead

val funlist :property list semval = 
	commasep function			>> map Pure
	

val impossible_line :property list semval = 
	getLit T_IMPOSSIBLE #-- eventlist 

val const_line :property list semval = 
	getLit T_CONST #-- constlist

val liveness_line :property list semval = 
	getLit T_DEAD #-- varlist 
	
val purity_line :property list semval = 
	getLit T_PURE #-- funlist 

val property :property list semval = 
	(   impossible_line 
	 || const_line 
	 || liveness_line 
	 || purity_line ) --# getLit T_SEMICOLON

val properties :property list semval = 
	repeat0 property			>> List.concat

val extension :string list semval = 
	(getLit T_RESTRICTS || getLit T_COLON)#-- namelist || success []

val ibody :property list semval = 
	getLit T_LBR #-- properties --# getLit T_RBR

val restriction : restriction semval = 
	getLit T_RESTRICTION #-- getName -- extension -- ibody
						>> (fn((n,e),p)=>(n,e,p))

val rbindings :block list semval = 
	   getName -- commasep getName		>> (fn (n,lst) => map (fn s=> 
							IBinding(n,s)) lst)	
val block :block list semval =
	   restriction				>> (fn s => [Inter s]) 
	|| rbindings 

val parser :(token,rl_ast) parser = 
	repeat0 (block --# getLit T_SEMICOLON) --# eof() >> List.concat


(* external access *)

fun scanFile fname = (* use this hook externally only for testing *)
	let val h = TextIO.openIn fname
	    val s = TextIO.inputAll h
	    val _ = TextIO.closeIn h
	in scanString tokenizer s end handle e => intErr e 200305081

(* FIXME: join two parsers in lazy manner and do the same with input stream? *)

fun parseFile fname = 
	scanList parser (valOf (scanFile fname))
	handle Option.Option => NONE

end
