(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/util/Err.sml,v
   Comment: Error handling for SCOPE
   Release: scope-0-16_final 

     Copyright (C) 2001-2003 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   WARNING: Error reporting is independent of Say machinery (BAD). 
   Just remember that all warnings and error messages are beyond verbosity
   control.

   TODO: Assertions are always evaluated (eager) despite the lazy interface.
         
   Err.sml,v 1.8 2003/12/15 16:03:10 wasowski Exp
*)

structure Err :> Err = struct

  open Say

  structure L   = Location
  type Location = L.Location

  (* are we running in debug mode? *)

  val release :bool ref = ref true

  fun release_mode true  = release := true
    | release_mode false = release := false

  (* is this guy really needed ?? *)

  datatype excty = Lexical      | Syntax   | Semant    | Assertion
                 | NotSupported | Internal | DebugInfo | Runtime
		 | WrongInput

  type exn_info = excty * Location * string option * string 
					(* type, pos, fname, msg *)
  exception ScopeExn of exn_info

  fun lexicalErr loc msg = raise ScopeExn (Lexical,loc, NONE,msg)
  fun syntaxErr  loc path msg = raise ScopeExn (Syntax, loc, path, msg)
  
  fun semErr msg = 
	raise ScopeExn (Semant, L.nilLocation, NONE,msg)
  
  fun runtimeErr msg =
	raise ScopeExn (Runtime, L.nilLocation, NONE,msg)
  
  fun wrongInput msg =
	raise ScopeExn (WrongInput, L.nilLocation, NONE, msg)
  
  fun warning msg = 
	IfVWarning(fn () => L.errPrompt ("Warning: " ^ msg ^ "\n"))

  
  
  fun impossibleL l "" = raise ScopeExn (Internal,l,NONE,"no description")
    | impossibleL l s  = raise ScopeExn (Internal,l,NONE,s)
    
  fun impossible s = impossibleL L.nilLocation s


  fun intErr (ScopeExn (Internal,b,c,d)) s = 
		raise ScopeExn (Internal,b,c,"#" ^ Int.toString s ^ " " ^ d)
    | intErr (ScopeExn e)  s = raise ScopeExn e
    | intErr (Fail str)    s = impossible ("#"^ Int.toString s ^ "#Fail:" ^ str)
    | intErr Empty         s = impossible ("#"^ Int.toString s ^ "#Empty"      )
    | intErr Subscript     s = impossible ("#"^ Int.toString s ^ "#Subscript"  )
    | intErr Match         s = impossible ("#"^ Int.toString s ^ "#Match"      )
    | intErr Option.Option s = impossible ("#"^ Int.toString s ^ "#Option"     )
    | intErr e             s = impossible ("#"^ Int.toString s ^ "#UnknownExc" )

  fun assert cond msg = 
	(!release) orelse cond() 
		orelse raise ScopeExn (Assertion,L.nilLocation,NONE,msg)


  fun addFile fname (t,l,_,msg) = (t,l,SOME fname,msg)

  fun finalize release e = ( release(); raise e )

  fun createLexerFile fpath =
	let val is = Nonstdio.open_in_bin fpath
	    val lexbuf = Lexing.createLexer 
		(fn buf => fn n => Nonstdio.buff_input is buf 0 n) 
	in (is,lexbuf) end
	
  fun displayLocation (loc, SOME fpath ) = 
	let val (is,lexbuf) = createLexerFile fpath
	in  L.errLocation (fpath,is,lexbuf) loc 
	    handle e => BasicIO.close_in is 
	end 
    | displayLocation _ = ()

  fun displayError (l,f,msg) = 
	let val _ = displayLocation (l,f) 
	in L.errPrompt (msg ^ "\n") end

  fun exnhandler (ScopeExn(t,l,f,msg)) retval = 
	let val pre = case t of
			  Lexical      => "Lexical error: "  
			| Syntax       => "Syntax error: "   
			| Semant       => "Static correctness error: " 
			| Runtime      => "Runtime error: "
			| NotSupported => "Unsupported feature: "
			| Internal     => "Internal error: "
			| DebugInfo    => "DebugInfo: "    
			| Assertion    => "Assertion failed: "
			| WrongInput   => "Error in input: "
	    val _ = displayError (l,f,pre ^ msg)
	in retval() end
    | exnhandler e retval = 
		(displayError (L.nilLocation,NONE,"Foreign exception."); 
		 raise e)

fun exnhandler1 e = exnhandler e (fn ()=>raise (Fail ""))
fun exnhandler0 e = exnhandler e (fn ()=>())

end
