(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/wuddy/wuddy-fn.sml,v
   Comment: Wuddy - slightly more abstract interface to Muddy (BuDDy)
   Release: scope-0-16_final
     
     Copyright (C) 2001-2004 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   NOTES:

   This generalization of muddy allows you to work with more abstract
   varialbes than integer indices of Muddy/Buddy. This is a trivial
   encapsulation of Muddy's structure. You still have to provide the variable
   numbering scheme in terms of two translating functions fromInt and toInt.

   This module can only work with one model at a time. Do not be
   misled by the functor! Only one instance of the package may work at
   a time! This is restriction of BUDDY, which without special
   cleverness can only work on one problem at a time (there is a
   single variable pool).

   You can still work on problems over different types in the same program,
   but either you have to do it sequentially (not with 2 at the same type) or
   by ensuring safe variable numbering.

   THIS MODULE USES IMPERATIVE FEATURES!!!

   wuddy-fn.sml,v 1.19 2004/05/13 12:55:34 wasowski Exp
*)

(* assumptions: *)
(* toInt should raise NotFound on variables outside the domain *)
(* toInt o fromInt = \x. x *)

functor Wuddy ( P : sig eqtype var 
			val toInt    : var -> int
			val toString : var -> string (* only for debugging *)
			val fromInt  : int -> var
		end ) = 

struct

    structure BU = BDDUtil

    (* exception safe versions of toInt toString and fromInt *)

    fun toInt    v = P.toInt    v handle NotFound => raise Fail 
			("Exception: reference to variable \"" ^ P.toString v ^
			 "\" not declared in bdd space.")
    fun toString v = P.toString v handle NotFound => raise Fail
			"Exception: probably reference to a variable not\
			\ declared in bdd space."
    fun fromInt  i = P.fromInt  i handle NotFound => raise Fail
			("Exception: reference to a variable \"" ^
			 Int.toString i ^ "\" not declared in bdd space.")

    type bdd = bdd.bdd
    type var = P.var

    type varnum = bdd.varnum (* should be non-negative *)

    val init      = bdd.init
    
    val setVarnum = bdd.setVarnum
    val getVarnum = bdd.getVarnum

    fun realloc (n :int) :unit = 
	if getVarnum() < n then setVarnum n else ()

    val isRunning = bdd.isRunning
    val done      = bdd.done

    (* nasty and slow. deservers sthg better *)
    fun listAllVars () = 
	let val ns = List.tabulate(getVarnum(), fn x => x)  
	in foldl (fn (x,e) => (fromInt x::e handle Domain => e)) [] ns end 

    val toBool   = bdd.toBool 
    val fromBool = bdd.fromBool

    val TRUE     = bdd.TRUE
    val FALSE    = bdd.FALSE 
    val equal    = bdd.equal
    val hash     = bdd.hash  

    fun always r = equal r TRUE
    fun never  r = equal r FALSE

    fun bddVar  (v :var) :bdd = bdd.ithvar ( toInt v )
    fun bddNVar (v :var) :bdd = bdd.nithvar ( toInt v )

    fun var (b :bdd) :var = fromInt (bdd.var b)

    val low  = bdd.low
    val high = bdd.high
    
    datatype bddop =
	And | Xor | Or | Nand | Nor | Imp | Biimp
      | Diff | Lessth | Invimp

    (* varSets *)

    type varSet = bdd.varSet

    fun makeset (vs :var list) :varSet = bdd.makeset (map toInt vs)
    fun makeset_ (vs :var vector) :varSet = bdd.makeset_ (Vector.map toInt vs)

    fun scanset  (vs :varSet) :var vector =
			Vector.map fromInt (bdd.scanset vs)

    val fromSet = bdd.fromSet
    val toSet_  = bdd.toSet_

    val support = bdd.support

    val apply  = bdd.apply
    val exist  = bdd.exist
    val forall = bdd.forall
    val appex  = bdd.appex 
    val appall = bdd.appall

    val DIFF   = bdd.DIFF
    val IMP    = bdd.IMP 
    val LESSTH = bdd.LESSTH
    val BIIMP  = bdd.BIIMP
    val OR     = bdd.OR
    val INVIMP = bdd.INVIMP
    val NAND   = bdd.NAND
    val NOR    = bdd.NOR
    val AND    = bdd.AND
    val XOR    = bdd.XOR

    val NOT = bdd.NOT   

    val ITE = bdd.ITE

    (* substitutions *)

    type substitution = bdd.pairSet

    fun substitution ( vvs: (var * var) list ) :substitution =
		bdd.makepairSet (map (fn (a,b)=> (toInt a,toInt b)) vvs)

    val replace = bdd.replace

    fun compose (b1 : bdd) (b2 :bdd) (v :var) :bdd =
			bdd.compose b1 b2 (toInt v)

    (* assignments *)

    type assignment = bdd.assignment

    fun assignment (vbs :(var * bool) list) :assignment =
		bdd.assignment (map (fn (v,b)=>(toInt v,b)) vbs)

    val fromAssignment = bdd.fromAssignment
    
    fun getAssignment (a :assignment) :(var * bool) list =
		map (fn (i,b) => (fromInt i,b)) (bdd.getAssignment a)


    val restrict = bdd.restrict
    val satone   = bdd.satone

    val simplify = bdd.simplify

    type fixed = bdd.fixed

    val FIXED = bdd.FIXED
    val FREE  = bdd.FREE
    
    fun addvarblock  (u :varnum) (v :varnum) = bdd.addvarblock u v

    val clrvarblocks = bdd.clrvarblocks

    type method = bdd.method
    val WIN2         = bdd.WIN2
    val WIN2ITE      = bdd.WIN2ITE
    val SIFT         = bdd.SIFT
    val SIFTITE      = bdd.SIFTITE
    val RANDOM       = bdd.RANDOM
    val REORDER_NONE = bdd.REORDER_NONE

    (* TODO: I wonder if reordering changes the identifiers. I guess not....
    *)

    val reorder          = bdd.reorder
    val autoReorder      = bdd.autoReorder
    val autoReorderTimes = bdd.autoReorderTimes

    val getMethod = bdd.getMethod
    val getTimes  = bdd.getTimes

    val disableReorder = bdd.disableReorder
    val enableReorder  = bdd.enableReorder

    type level = int

    fun varToLevel (v :var) :level = bdd.varToLevel (toInt v)
    fun varAtLevel (l :level) :var = fromInt (bdd.varAtLevel l)

    val fnprintdot = bdd.fnprintdot
    val printdot   = bdd.printdot
    val fnprintset = bdd.fnprintset
    val printset   = bdd.printset

    val bddSave = bdd.bddSave
    val bddLoad = bdd.bddLoad

    val satcount  = bdd.satcount
   
    type nodetable = int * (var * int * int) vector
    fun nodetable (b :bdd) :nodetable =
		let val (n,vijs) = bdd.nodetable b
		in (n,Vector.map (fn (v,i,j)=> (fromInt v,i,j)) vijs ) end 

    val nodecount = bdd.nodecount

    val setMaxincrease = bdd.setMaxincrease
    val setCacheratio  = bdd.setCacheratio
    val verbosegc      = bdd.verbosegc

    val stats   = bdd.stats

    (* my add ons *)

    (* skip variables not in the domain *)
    
fun sprMapping () :string =
	let fun stl i = 
		let val s = explode (toString(fromInt i))
		in (implode o rev o List.take) (rev s,Int.min(15,length s)) end
	    fun pr (i,ss) = 
		(Int.toString i ^ " -> " ^ stl i ^ "\n" ::ss) handle _ => ss
	in String.concat (foldl pr [] (List.tabulate(getVarnum(), fn x=>x))) end 

fun prMapping () :unit = print (sprMapping())

(* show BDD in graphical form *)

fun show (b :bdd) :unit = (* FIXME UNSFAFE *)
	let val n = FileSys.tmpName();
	    val n1= n ^ ".ps"
	    val _ = fnprintdot n b;
	    val _ = Process.system ("dot -Tps " ^ n ^ " -o " ^ n1);
	    val _ = prMapping();
	    val _ = Process.system ("gv " ^ n1);
	in FileSys.remove n; FileSys.remove n1 end
   
 (* existentially quantify over everything not in the given set ss *)

 fun restrict_to (r :bdd) (vs :var list) :bdd =
	    BU.restrict_to r (map toInt vs)

    (* generalized n-argument xor : builds a formula
       which is true if exactly one of variables is true *)

  fun generalized_xor (vs :var list) :bdd = 
		BU.generalized_xor (map toInt vs)

  (* strong support stuff *)

  fun is_strong (bdd :bdd) (v :var) :bool = 
	BU.is_strong bdd (toInt v)

  (* restrict bdd only to not-strong part of support 
     returns the eliminated part of (strong) support *)

  val weaken_support = BU.weaken_support

end

(* Functor Wuddy is a generalization of interface of Muddy by Ken Friis
   Larsen and Jakob Lichtenberg (which in turn is a Standard ML interface to
   BuDDy BDD library by Joern Linds). Following help is copied form Muddy and
   as such is not up to date.

   Type [bdd] is the abstract type of BDDs of variables of type P.vartype

   [init nodesize cachesize] initiates the bdd package and must be
   called before any bdd operations are done.  nodesize is the initial
   number of nodes in the nodetable and cachesize is the fixed size of
   the internal caches.  Typical values for nodesize are 10000 nodes
   for small test examples and up to 1000000 nodes for large examples.
   A cache size of 10000 seems to work good even for large examples,
   but lesser values should do it for smaller examples.

   The initial number of nodes is not critical for any bdd operation
   as the table will be resized whenever there are too few nodes left
   after a garbage collection.  But it does have some impact on the
   efficency of the operations.

   init and done should only be called once per session.

   [setvarnum num] is used to define the number of variables used in
   the bdd package. It may be called more than one time, but only to
   increase the number of variables. num is the number of variables to
   use.

   [done ()] frees all memory used by the bdd package and resets the
   package to its initial state.

   [toBool r] return the boolean represented by r.  Raises Domain if r
   can not be represented as true or false.

   [fromBool b] gives the bdd representing b.

   [equal x y] is true if x and y represent equivalent boolean
   expressions. 

   [always r] returns true if r is a tautology (TRUE constant bdd).
   Syntactic sugar for "equal r TRUE".

   [never r] returns true if r is never true (contains a contradiction).
   Syntactic sugar for "equal r FALSE".
   

   [ithvar i] gives the bdd representing the i'th variable.  The
   requested variable must be in the range define by setvarnum
   starting with 0 being the first.

   [nithvar i] gives the bdd representing the negation of the i'th
   variable.  The requested variable must be in the range defined by
   setvarnum starting with 0 being the first.

   [var r] gets the variable labeling r.

   [low r] gets the false branch of r.
   
   [high r] gets the true branch of r.

   Type [varSet] is an effective representation of sets of variables.

   [makeset varlist] makes the varSet with the elements in varlist.
   There is no constraint on varlist: it may be empty, contain duplets,
   or negative numbers.  Duplets and negative numbers will just be
   filtered out.

   [makeset_ varvector] makes the varSet with the elements in
   varvector.  There are many constraints on varvector: it must be
   sorted in increasing order, it may not contain duplets nor negative
   numbers.  This function is more effictive than makeset, but it is
   highly unsafe and you should only use it when you can guarantee
   that the input satisfies the constraints.  If it doesn't satisfy the
   constraints "weird things" may happen, and your ML program may even
   "core dump"!

   [scanset varset] gives the vector of the variables in varset.

   [fromSet varset] gives the bdd representing the conjunction of
   all the variables in varset.

   [toSet_ r] converts the bdd r to a varSet; no checks are performed
   to check that r really represents a varSet therefore it should be
   used with care.

   [support r] gives the set of variables that r depends on.

   Type [bddop] represent the binary boolean operator which can be used
   when constructing bdds. Truth tables:

   x y  |  And  Xor  Or  Nand  Nor  Imp  Biimp  Diff  Lessth  Invimp
  ------|------------------------------------------------------------
   T T  |   T    F   T    F     F    T     T     F      F       T
   T F  |   F    T   T	  T     F    F     F     T      F       T
   F T  |   F    T   T	  T     F    T     F     F      T       F
   F F  |   F    F   F	  T     T    T     T     F      F       T

   [apply x y opr] constructs the bdd using the binary boolean
   operator.

   [exist varset r] constructs the existential quantification over all
   the variables in the varset of r.

   [forall varset r] constructs the universal quantification over all
   the variables in the varset of r.

   [appex x y opr varset] equivalent to 'exist varset (apply x y opr)'
   but more efficient.

   [appall x y opr varset] equivalent to 'forall varset (apply x y opr)'
   but more efficient.

   [DIFF(x,y)] equivalent to 'apply x y Diff', but this is nice for
   infix.

   [IMP(x,y)] equivalent to 'apply x y Imp', but this is nice for infix.
    
   [LESSTH(x,y)] equivalent to 'apply x y Lessth', but this is nice for
   infix.
    
   [BIIMP(x,y)] equivalent to 'apply x y Biimp', but this is nice for
   infix.

   [OR(x,y)] equivalent to 'apply x y Or', but this is nice for infix.

   [INVIMP(x,y)] equivalent to 'apply x y Invimp', but this is nice for
   infix.

   [NAND(x,y)] equivalent to 'apply x y Nand', but this is nice for infix.

   [NOR(x,y)] equivalent to 'apply x y Nor', but this is nice for infix.

   [AND(x,y)] equivalent to 'apply x y And', but this is nice for infix.

   [XOR(x,y)] equivalent to 'apply x y Xor', but this is nice for infix.

   [NOT r] construct the negation of r.

   [ITE x y z] equivalent to 'OR(AND(x, y), AND(NOT x, z))'; but more
   efficient.

   Type [pairSet] is used to represet substitutions of variables with
   other variables.

   [makepairSet [(x,x'),(y,y'),...,(z,z')]] creates the substitution
   which substitute x' for x, y' for y, ..., and z' for z.

   [replace r pairset] perfoms the substitution pairset on r.

   [compose r1 r2 var] substitute r2 for var in r1.

   Type [assignment] represents an assignment of variables. The
   assignment [(x1,true), (x3,false), (x2, true)] corresponds to the
   bdd 'AND(x1,AND(NOT x3, x2))'.

   [restrict r assign] ... restrict the variables in assign to TRUE or
   FALSE in r.  [documentation not complete here].

   [satone r] finds a satisfying variable assignment.

   [simplify r dom] tries to simplify r by restricting it to domain d,
   ie. 'r AND dom  =  dom AND (simplify r dom)'.
   
   !!!! DOCUMENTATION IS MISSING ON VARIABLE REORDER STUFF !!!!

   [disableReorder()] diaples automatic reordering.  Reordering is
   enabled by default as soon as ant variable blocks have been
   defined.

   [enableReorder()] Re-enables reordering after a call to
   disableReorder.

   [satcount r] returns how many possible variable assignments there
   exist such that r is satisfied, taking all defined variables into
   account.

   [printdot r] prints r in a format suitable for use by the graph drawing
   program dot, on standard output 
   (dot can be obtained from http://www.research.att.com/sw/tools/graphviz/)

   [fnprintdot fname r] prints r in a format suitable for use by the graph
   drawing program dot in the file fname.

   [show r] Calls fprintdot to generate temporary file. Then translates it to
   postscript using dot and displays in gv. Both dot and gv have to be
   installed and visible in the path.

   [printset r] prints all the truth assignments for r that would
   satisfy r, on standard output.

   [fnprintset fname r] prints all the truth assignments for r that would
   satisfy r, in the file fname.

   [bddSave name r] saves r in the file name.

   [bddLoad name] returns the bdd saved in the file name.

   [nodetable r] returns the nodetable for r.

   [setMaxincrease n] tells BuDDy that the maximum of new nodes added
   when doing an expansion of the nodetable should be n.  Return the
   old maximum.

   [setCacheratio n] sets the cache ratio to n.  For example, if n is
   four then the internal caches will be 1/4th the size of the
   nodetable.

   [verbosegc(SOME(pregc,postgc))] instructs BuDDy to print
   pregc when a BuDDy GC is initiated and print postgc when the
   BuDDy GC is completed.

   [stats()] gives various statstistical information from the
   underlying C library:
     produced     : total number of new nodes ever produced.
     nodenum      : currently allocated number of bdd nodes.
     maxnodenum   : user defined maximum number of bdd nodes.
     freenodes    : number of currently free nodes.
     minfreenodes : minimum number of nodes that should be left after
                    a garbage collection (in the bdd library).
     varnum       : number of defined bdd variables,
     cachesize    : number of entries in the internal caches.
     gbcnum       : number of garbage collection done in the bdd library.

   [is_strong bdd v] returns true if v belongs to strong support of bdd, i.e
   i.e. it has the same assignment in all satisfying assignment (either
   always true or always false).

*)
